#ifndef SIMULATOR_HPP
#define SIMULATOR_HPP

#include <limits>
#include <algorithm>

#include "Trace.hpp"

class Simulator
{
public:
	Simulator(int n=2, int k=2, 
		int P=1, double W=0.0, 
		double C=0.0, double V=0.0, double R=0.0, double D=0.0, 
		double mu_s=INF, double mu_f=INF, double alpha=0.0) :
		n(n), k(k), P(P),
		W(W), C(C), V(V), R(R), D(D),
		mu_s(mu_s), mu_f(mu_f),
		alpha(alpha),
		Pgroup(0),
		H(0.0),
		nR(0.0), nC(0.0), nE(0.0)
		{}

	// All time variables are in seconds

	// Input
	int n, k, P;
	double W, C, V, R, D; // Pattern length, Checkpoint, Recovery and Dowtime costs
	double mu_s, mu_f;
	double alpha;

	int Pgroup;

	// Output
	double H; 		// Overhead
	double Hopt;
	double nR, nC, nE;		// Number of recoveries, checkpoints and verifications

private:
	double cTime, cCPU;

public:

	void runOneReplica(Trace &failstop, Trace &silent, double &lost, bool &F, bool &S)
	{
		double serr = silent.next(cCPU);
		double ferr = failstop.next(cTime);

		// If there is a fail-stop error
		if(ferr - cTime < W)
		{
			F = true;
			S = false;
			lost = ferr - cTime;
		}
		// Else if there is a silent error
		else if(serr - cCPU < W)
		{
			F = false;
			S = true;
			lost = 0;
		}
		else
		{
			lost = 0;
			F = false;
			S = false;
		}
	}

	void runOneProcess(Trace *failstop, Trace *silent, double &lost, bool &F, bool &S)
	{
		std::vector<double> tlost;
		int failedi = 0;
		int corruptedi = 0;

		lost = 0;
		F = S = false;

		for(int i=0; i<n; i++)
		{
			bool Fi, Si;
			double lost;

			runOneReplica(failstop[i], silent[i], lost, Fi, Si);

			if(Fi)
			{
				failedi++;
				tlost.push_back(lost);
			}
			else if(Si)
				corruptedi++;
		}

		// Failstop
		if(n - failedi < k)
		{
			std::sort(tlost.begin(), tlost.end()); 
			lost = tlost[k-1];
			F = true;
		}
		// Silent
		else if(n - failedi - corruptedi < k)
			S = true;

		nE += corruptedi;
		if(n - failedi < k) 
			nE += n - k + 1;
		else
			nE += failedi;
	}

	// run will simulate the execution of nw patterns of size W
	void run(const int nw)
	{
		if(k > n)
		{
			printf("ERROR: k must be smaller than or equal to n.\n");
			return;
		}

		nR = 0;
		nC = 0;
		nE = 0;

		cTime = 0.0;
		cCPU = 0.0;

		bool F, S, failure, error;
		double minlost = 0;
		double lost = 0;
		
		std::vector<Trace> failstop(P*n);
		std::vector<Trace> silent(P*n);
		for(int i=0; i<P*n; i++)
		{
			failstop[i].mu = mu_f;
			silent[i].mu = mu_s;
		}

		int exec = 0;
		while(exec < nw)
		{
			if(W == 0) break;

			goto pattern;

			disk_recovery:
				nR++;
				cTime += R;
				goto pattern;

			pattern:
				lost = 0;
				minlost = INF;
				failure = error = false;

				for(int i=0; i<P; i++)
				{
					runOneProcess(&failstop[i*n], &silent[i*n], lost, F, S);

					if(F)
					{
						failure = true;
						if(lost < minlost)
							minlost = lost;
					}
					else if(S)
						error = true;
				}

				// Failstop > 0
				if(failure)
				{
					cTime += minlost + D;
					cCPU += minlost;
					goto disk_recovery;
				}
				else if(error)
				{
					cTime += W + V;
					cCPU += W;
					goto disk_recovery;
				}
				else
					cTime += W + V;

//			checkpoint:
			cTime += C;
			exec++;
			nC++;
		}

		// Done!
		double speedup = alpha + (1-alpha)/P;
		
		if(Pgroup > 1)
			speedup =  alpha + (1-alpha)/Pgroup;

		double totaltime = cTime * speedup;
		double totalworktime = nw * W;

		// Compute nR per day
		nR /= (cTime / (24.0*3600));
		nE /= (cTime / (24.0*3600));

		// Normalized overhead with respect to the total worktime
		H = totaltime / totalworktime;
	}

private:
};

#endif
