#include <cstdio>
#include <cstdlib>
#include <time.h>
#include <sys/time.h>
#include <cstring>
#include <string>
#include <omp.h>

#include "Simulator.hpp"

int fact(int n){ int f = 1; for(int i = 1; i <= n; i++) f *= i; return f; }

double choose(int n, int k){ return fact(n) / (fact(k) * fact(n-k)); }

double process_Topt(const Simulator &sim)
{   
	int n = sim.n;
	int k = sim.k;
	int P = sim.P;
	double C = sim.C;
	double V = sim.V;
	double LA = 1.0/sim.mu_s + 1.0/sim.mu_f;
	double la_f = 1.0/sim.mu_f;

	if(n == 2 && k == 2)
		return pow((V+C) / ((2*LA - la_f) * P), 1.0/2.0);
	else if(n ==3 && k ==2)
		return pow((V+C) / ((6*LA*LA - 2*la_f*la_f) * P), 1.0/3.0);
	else
	{
		printf("Not Implemented.\n");
		return -1;
	}
}

double process_Popt(const Simulator &sim, double c)
{
	int n = sim.n;
	int k = sim.k;

	double LA = 1.0/sim.mu_s + 1.0/sim.mu_f;
	double la_f = 1.0/sim.mu_f;
	double alpha = sim.alpha;

	if(n == 2 && k == 2)
		return pow(pow((1-alpha)/alpha, 2) * (1.0 / ((2*LA - la_f) * c)), 1.0/3.0);
	else if(n ==3 && k ==2)
		return pow(pow((1-alpha)/alpha, 3) * (4.0 / ((3*LA*LA - la_f*la_f) * c*c)), 1.0/4.0);
	else
	{
		printf("Not Implemented.\n");
		return -1;
	}
}

double group_Popt(const Simulator &sim, double c)
{
	int n = sim.n;
	int k = sim.k;
	double LA = 1.0/sim.mu_s + 1.0/sim.mu_f;
	double la_f = 1.0/sim.mu_f;
	double alpha = sim.alpha;

	double aa = (1-alpha)/alpha;

	if(n == 2 && k == 2)
		return pow(pow(aa, 2) * (1.0 / ((2*LA - la_f) * c)), 1.0/3.0);
	else if(n ==3 && k ==2)
		return pow(pow(aa, 3) * (1.0 / ((6*LA*LA - la_f*la_f) * c*c)), 1.0/5.0);
	else
	{
		printf("Not Implemented.\n");
		return -1;
	}
}

double process_Hopt(const Simulator &sim)
{
	int n = sim.n;
	int k = sim.k;
	int P = sim.P;
	//double T = sim.W;
	double C = sim.C;
	double V = sim.V;
	double LA = 1.0/sim.mu_s + 1.0/sim.mu_f;
	double la_f = 1.0/sim.mu_f;

	double speedup = 1.0/(sim.alpha + (1.0-sim.alpha)/P);
	
	if(sim.Pgroup > 1)
		speedup = 1.0/(sim.alpha + (1.0-sim.alpha)/sim.Pgroup);

	if(n == 2 && k == 2)
	{
		//double a = speedup / (1 + 2 * pow( (2*LA - la_f) * (V+C)*P, 1.0/2.0));
		//printf("Popt(2) = %f\n", a);
		return speedup / (1 + 2 * pow( (2*LA - la_f) * (V+C)*P, 1.0/2.0));
	}
	else if(n == 3 && k == 2)
	{
		//double a = speedup / (1 + 3 * pow( ((3*LA*LA - la_f*la_f)/4.0) * pow((V+C),2) * P, 1.0/3.0));
		//printf("Popt(3) = %f\n", a);
		return speedup / (1 + 3 * pow( ((3*LA*LA - la_f*la_f)/4.0) * pow((V+C),2) * P, 1.0/3.0));
	}

	return -1;
}

class Result
{
public:
	int Popt;
	double Topt, Hopt, Hsim;
	double nC, nR, nE;
	int Q, n, k;

	Result() : Popt(0), Topt(0), Hopt(0), Hsim(0), nC(0), nR(0), nE(0), Q(0), n(0), k(0){}

	void print(FILE *out)
	{
		double dAbs = fabs(1.0/(Hsim*Q) - Hopt/Q);
		double dRel = fabs(1.0/(Hsim*Q) - (Hopt/Q)/(1.0/(Hsim*Q)));

		printf("Q=%d n=%d k=%d P=%d Tt=%.0f Et=%.2f Es=%.2f dAbs=%.2f dRel=%.2f nC=%.2f, nR=%.2f, nE=%.2f\n", 
			Q, n, k, Popt, Topt, Hopt/Q, 1.0/(Hsim*Q), dAbs, dRel, nC, nR, nE);

		fprintf(out, "%d %d %d %d %e %e %e %e %e %e %e %e\n", 
			Q, n, k, Popt, Topt, Hopt/Q, 1.0/(Hsim*Q), dAbs, dRel, nC, nR, nE);
	}
};

Result average(Simulator &sim, int Q, int nruns, int nw)
{
	Result r;
	#pragma omp parallel shared(r)
  	{
  		struct timeval time;
		gettimeofday(&time, NULL);
		srand(((time.tv_sec * 1000) + (time.tv_usec / 1000)) ^ omp_get_thread_num());
		#pragma omp for
		for(int i=0; i<nruns; i++)
		{
			Simulator simc = sim;
			simc.run(nw);

			#pragma omp critical
			{
				r.nC += simc.nC;
				r.nR += simc.nR;
				r.nE += simc.nE;
				r.Hsim += simc.H;
			}
		}
	}

	r.nC /= nruns;
	r.nR /= nruns;
	r.nE /= nruns;
	r.Hsim /= nruns;

	r.Q = Q;
	r.n = sim.n;
	r.k = sim.k;
	r.Popt = sim.P;

	if(sim.Pgroup > 1)
		r.Popt = sim.Pgroup;

	r.Topt = sim.W;

	r.Hopt = sim.Hopt;
	
	return r;
}

Simulator createProcSim(int Q, double P, double W, int n, int k, double d, double c, double v, double mu_s, double mu_f, double alpha)
{
	Simulator sim(n, k, P, W, c, v, c, 0.0, mu_s, mu_f, alpha);
	sim.C = c + d/(Q/n);
	sim.V = v * sim.C;

	sim.P = (P != -1)? P : std::min(floor((double)Q/n), floor(process_Popt(sim,c)));
	sim.W = (W != -1)? W : process_Topt(sim);

	sim.Hopt = process_Hopt(sim);

	return sim;
}

Simulator createGroupSim(int Q, double P, double W, int n, int k, double d, double c, double v, double mu_s, double mu_f, double alpha)
{
	Simulator sim(n, k, P, W, c, v, c, 0.0, mu_s, mu_f, alpha);
	sim.C = c + d/((double)Q/n);
	sim.V = v * sim.C;

	sim.P = 1;
	sim.Pgroup = (P != -1)? P : std::min(floor((double)Q/n), floor(group_Popt(sim,c)));

	sim.mu_s = mu_s / sim.Pgroup;
	sim.mu_f = mu_f / sim.Pgroup;

	sim.W = (W != -1)? W : process_Topt(sim);

	sim.Hopt = process_Hopt(sim);

	return sim;
}

void procExperiment(FILE *file, int Q, int P, double W, int n, int k, double d, double c, double v, double mu_s, double mu_f, double alpha, int nruns, int nw)
{
	Simulator sim = createProcSim(Q, P, W, n, k, d, c, v, mu_s, mu_f, alpha);
	average(sim, Q, nruns, nw).print(file);
}

void groupExperiment(FILE *file, int Q, int P, double W, int n, int k, double d, double c, double v, double mu_s, double mu_f, double alpha, int nruns, int nw)
{

	Simulator sim = createGroupSim(Q, P, W, n, k, d, c, v, mu_s, mu_f, alpha);
	average(sim, Q, nruns, nw).print(file);
}

void create(const char *name, FILE **fp22, FILE **fp32, FILE **fg22, FILE **fg32)
{
	char p22[256], p32[256], g22[256], g32[256];
	sprintf(p22, "results/proc/var_%s_n2_k2.out", name);
	sprintf(p32, "results/proc/var_%s_n3_k2.out", name);
	sprintf(g22, "results/group/var_%s_n2_k2.out", name);
	sprintf(g32, "results/group/var_%s_n3_k2.out", name);

	if((*fp22 = fopen(p22, "w")) <= 0) { printf("ERROR\n"); fclose(*fp22); }
	//if((*fp32 = fopen(p32, "w")) <= 0) { printf("ERROR\n"); fclose(*fp32); }
	if((*fg22 = fopen(g22, "w")) <= 0) { printf("ERROR\n"); fclose(*fg22); }
	if((*fg32 = fopen(g32, "w")) <= 0) { printf("ERROR\n"); fclose(*fg32); }
}

void close(FILE *fp22, FILE *fp32, FILE *fg22, FILE *fg32)
{
	if(fp22) fclose(fp22);
	//if(fp32) fclose(fp32);
	if(fg22) fclose(fg22);
	if(fg32) fclose(fg32);
}

void runall(FILE *fp22, FILE *fp32, FILE *fg22, FILE *fg32,
	int Q, int P, double W, double d, double c, double v, double mu_s, double mu_f, double alpha, int nruns, int nw)
{
	//procExperiment(fp22, Q, 2, 2, P, W, c, C, V, mu_s, alpha, nruns, nw);
	//procExperiment(fp32, Q, std::min((Q/3.0),(double)P), W, 3, 2, d, c, v, mu_s, mu_f, alpha, nruns, nw);
	groupExperiment(fg22, Q, std::min((Q/2.0),(double)P), W, 2, 2, d, c, v, mu_s, mu_f, alpha, nruns, nw);
	groupExperiment(fg32, Q, std::min((Q/3.0),(double)P), W, 3, 2, d, c, v, mu_s, mu_f, alpha, nruns, nw);
}

int main(void)
{
	int nw = 4096;
	int nruns = 1024;

	FILE *fp22, *fp32, *fg22, *fg32;

	double Q = 1e6;
	double v = 0;
	double alpha = 1e-6;

	double mu = 1e3;
	double LA = 1/mu;

	// mu = 1e6;
	// create("FS-Mu-fixed-1e6", &fp22, &fp32, &fg22, &fg32);
	// for(double muf=1e6; muf>=1e2; muf/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, mu*Q, muf*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// mu = 1e5;
	// create("FS-Mu-fixed-1e5", &fp22, &fp32, &fg22, &fg32);
	// for(double muf=1e6; muf>=1e2; muf/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, mu*Q, muf*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// mu = 1e4;
	// create("FS-Mu-fixed-1e4", &fp22, &fp32, &fg22, &fg32);
	// for(double muf=1e6; muf>=1e2; muf/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, mu*Q, muf*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// mu = 1e3;
	// create("FS-Mu-fixed-1e3", &fp22, &fp32, &fg22, &fg32);
	// for(double muf=1e6; muf>=1e2; muf/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, mu*Q, muf*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);


	// mu = 1e6;
	// LA = 1/mu;
	// create("FS-Mu-1e6", &fp22, &fp32, &fg22, &fg32);
	// for(double eps=0; eps<=1; eps+=0.1)
	// {
	// 	double muS = 1/(LA*(1-eps));
	// 	double muF = 1/(LA*eps);
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, muS*Q, muF*Q, alpha, nruns, nw);	
	// }
	// close(fp22, fp32, fg22, fg32);

	// mu = 1e5;
	// LA = 1/mu;
	// create("FS-Mu-1e5", &fp22, &fp32, &fg22, &fg32);
	// for(double eps=0; eps<=1; eps+=0.1)
	// {
	// 	double muS = 1/(LA*(1-eps));
	// 	double muF = 1/(LA*eps);
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, muS*Q, muF*Q, alpha, nruns, nw);	
	// }
	// close(fp22, fp32, fg22, fg32);

	mu = 1e3;
	LA = 1/mu;
	create("FS-Mu-1e3", &fp22, &fp32, &fg22, &fg32);
	for(double eps=0; eps<=1; eps+=0.1)
	{
		double muS = 1/(LA*(1-eps));
		double muF = 1/(LA*eps);
		runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, muS*Q, muF*Q, alpha, nruns, nw);	
	}
	close(fp22, fp32, fg22, fg32);

	// mu = 1e3;
	// LA = 1/mu;
	// create("FS-Mu-1e3", &fp22, &fp32, &fg22, &fg32);
	// for(double eps=0; eps<=1; eps+=0.1)
	// {
	// 	double muS = 1/(LA*(1-eps));
	// 	double muF = 1/(LA*eps);
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, muS*Q, muF*Q, alpha, nruns, nw);	
	// }
	// close(fp22, fp32, fg22, fg32);

	return 0;
}
