#ifndef SIMULATOR_HPP
#define SIMULATOR_HPP

#include <limits>

#include "Trace.hpp"

class Simulator
{
public:
	Simulator(int n=2, int k=2, 
		int P=1, double W=0.0, 
		double C=0.0, double V=0.0, double R=0.0, double D=0.0, 
		double mu_f=INF, double mu_s=INF, double alpha=0.0,
		unsigned int seed = 0) :
		n(n), k(k), P(P),
		W(W), C(C), V(V), R(R), D(D),
		mu_f(mu_f), mu_s(mu_s),
		alpha(alpha),
		Pgroup(0),
		H(0.0),
		nR(0.0), nC(0.0),
		seed(seed)
		{}

	// All time variables are in seconds

	// Input
	int n, k, P;
	double W, C, V, R, D; // Pattern length, Checkpoint, Recovery and Dowtime costs
	double mu_f, mu_s;
	double alpha;

	int Pgroup;

	// Output
	double H; 		// Overhead
	double Hopt;
	double nR, nC, nE;		// Number of recoveries, checkpoints and verifications

	unsigned int seed;

private:
	double cTime, cCPU;

public:

	void runOneReplica(Trace &failstop, Trace &silent, double &lost, bool &F, bool &S)
	{
		double serr = silent.next(cCPU);
		double ferr = failstop.next(cTime);

		// If there is a fail-stop error
		if(ferr - cTime < W)
		{
			F = true;
			S = false;
			lost = ferr - cTime;
		}
		// Else if there is a silent error
		else if(serr - cCPU < W)
		{
			F = false;
			S = true;
			lost = 0;
		}
		else
		{
			lost = 0;
			F = false;
			S = false;
		}
	}

	void runOneProcess(Trace *failstop, Trace *silent, double &lost, bool &F, bool &S)
	{
		int alivei = 0;
		int corruptedi = 0;
		double maxlost = 0;
		for(int i=0; i<n; i++)
		{
			bool Fi, Si;
			double lost;

			runOneReplica(failstop[i], silent[i], lost, Fi, Si);

			if(Fi)
			{
				if(lost > maxlost)
					maxlost = lost;
			}
			else
			{
				alivei++;
				if(Si) corruptedi++;
			}
		}

		// Failstop
		if(alivei == 0)
		{
			lost = maxlost;
			F = true;
			S = false;
		}
		// Silent
		else if(alivei - corruptedi < k)
		{
			lost = 0;
			F = false;
			S = true;
		}
		else
		{
			lost = 0;
			F = false;
			S = false;
		}

		nE += corruptedi;
	}

	// run will simulate the execution of nw patterns of size W
	void run(const int nw)
	{
		if(k > n)
		{
			printf("ERROR: k must be smaller than or equal to n.\n");
			return;
		}

		nR = 0;
		nC = 0;
		nE = 0;

		cTime = 0.0;
		cCPU = 0.0;

		bool F, S;
		int alive = 0;
		int corrupted = 0;
		double minlost = 0;
		double lost = 0;
		
		std::vector<Trace> failstop(P*n);
		std::vector<Trace> silent(P*n);
		for(int i=0; i<P*n; i++)
		{
			failstop[i].seed = seed+i;
			silent[i].seed = seed+P*n+i;
			failstop[i].mu = mu_f;
			silent[i].mu = mu_s;
		}

		int exec = 0;
		while(exec < nw)
		{
			if(W == 0) break;

			goto pattern;

			disk_recovery:
				nR++;
				cTime += R;

			pattern:
				lost = 0;
				minlost = INF;
				alive = corrupted = 0;

				for(int i=0; i<P; i++)
				{
					runOneProcess(&failstop[i*n], &silent[i*n], lost, F, S);

					if(F)
					{
						if(lost < minlost)
							minlost = lost;
					}
					else
					{
						alive++;
						if(S) corrupted++;
					}
				}

				// Failstop > 0
				if(alive == 0)
				{
					cTime += minlost + D;
					cCPU += minlost;
					goto disk_recovery;
				}
				else
				{
					cTime += W + V;
					cCPU += W;
					// Silent > 0
					if(corrupted > 0)
						goto disk_recovery;
				}

//			checkpoint:
			cTime += C;
			exec++;
			nC++;
		}

		// Done!
		double speedup = alpha + (1-alpha)/P;
		
		if(Pgroup > 1)
			speedup =  alpha + (1-alpha)/Pgroup;

		double totaltime = cTime * speedup;
		double totalworktime = nw * W;

		// Compute nR per day
		nR /= (cTime / (24.0*3600));
		nE /= (cTime / (24.0*3600));

		// Normalized overhead with respect to the total worktime
		H = totaltime / totalworktime;
	}

private:
};

#endif
