#include <cstdio>
#include <cstdlib>
#include <time.h>
#include <sys/time.h>
#include <cstring>
#include <string>
#include <omp.h>

#include "Simulator.hpp"

int fact(int n){ int f = 1; for(int i = 1; i <= n; i++) f *= i; return f; }

double choose(int n, int k){ return fact(n) / (fact(k) * fact(n-k)); }

double process_Topt(const Simulator &sim)
{   
	int n = sim.n;
	int k = sim.k;
	int P = sim.P;
	double C = sim.C;
	double V = sim.V;
	double la = 1.0/sim.mu_s;

	double beta = choose(n,k-1)*(n-k+1);

	return pow((C+V)/(beta * P * pow(la, n-k+1)), 1.0/(n-k+2));
}

double process_Popt(const Simulator &sim, double c)
{
	int n = sim.n;
	int k = sim.k;
	double la = 1.0/sim.mu_s;
	double alpha = sim.alpha;

	double gamma = pow((n-k+1),n-k+1)/choose(n,k-1);

	return pow(gamma * pow((1-alpha)/alpha, n-k+2) * pow(1.0/(la * c), n-k+1), 1.0/(n-k+3));
}

double group_Popt(const Simulator &sim, double c)
{
	int n = sim.n;
	int k = sim.k;
	double la = 1.0/sim.mu_s;
	double alpha = sim.alpha;

	double beta = choose(n,k-1)*(n-k+1);

	return pow((1.0/beta) * pow((1-alpha)/alpha, n-k+2) * pow(1.0/(la * c), n-k+1), 1.0/(2*n-2*k+3));
}

double process_Hopt(const Simulator &sim)
{
	int n = sim.n;
	int k = sim.k;
	int P = sim.P;
	double T = sim.W;
	double C = sim.C;
	double V = sim.V;
	double la = 1.0/sim.mu_s;

	double speedup = sim.alpha + (1.0-sim.alpha)/sim.P;

	if(sim.Pgroup > 1)
		speedup = sim.alpha + (1.0-sim.alpha)/sim.Pgroup;

	return speedup * ((1.0 + choose(n,k-1) * P * pow(la*T,n-k+1) + (V+C)/T));
}

class Result
{
public:
	int Popt;
	double Topt, Hopt, Hsim;
	double nC, nR, nE;
	int Q, n, k;

	Result() : Popt(0), Topt(0), Hopt(0), Hsim(0), nC(0), nR(0), nE(0), Q(0), n(0), k(0){}

	void print(FILE *out)
	{
		double dAbs = fabs(1.0/(Hsim*Q) - 1.0/(Hopt*Q));
		double dRel = fabs((1.0/(Hsim*Q) - 1.0/(Hopt*Q))/(1.0/(Hsim*Q)));

		printf("Q=%d n=%d k=%d P=%d Tt=%.0f Et=%.2f Es=%.2f dAbs=%.2f dRel=%.2f nC=%.2f, nR=%.2f, nE=%.2f\n", 
			Q, n, k, Popt, Topt, 1.0/(Hopt*Q), 1.0/(Hsim*Q), dAbs, dRel, nC, nR, nE);

		fprintf(out, "%d %d %d %d %e %e %e %e %e %e %e %e\n", 
			Q, n, k, Popt, Topt, 1.0/(Hopt*Q), 1.0/(Hsim*Q), dAbs, dRel, nC, nR, nE);
	}
};

Result average(Simulator &sim, int Q, int nruns, int nw)
{
	unsigned int salt = rand();

	Result r;
	#pragma omp parallel for shared(r)
	for(int i=0; i<nruns; i++)
	{
		unsigned int seed = salt + i;
		Simulator simc = sim;
		simc.seed = rand_r(&seed);
		simc.run(nw);

		#pragma omp critical
		{
			r.nC += simc.nC;
			r.nR += simc.nR;
			r.nE += simc.nE;
			r.Hsim += simc.H;
		}
	}

	r.nC /= (double)(nruns * nw);
	r.nR /= nruns;
	r.nE /= nruns;
	r.Hsim /= nruns;

	r.Q = Q;
	r.n = sim.n;
	r.k = sim.k;
	r.Popt = sim.P;

	if(sim.Pgroup > 1)
		r.Popt = sim.Pgroup;

	r.Topt = sim.W;

	r.Hopt = sim.Hopt;
	
	return r;
}

Simulator createProcSim(int Q, int n, int k, double P, double W, double d, double c, double v, double mu_s, double alpha)
{
	Simulator sim(n, k, P, W, c, v, c, 0.0, INF, mu_s, alpha);
	sim.C = c + d/(Q/n);
	sim.V = v * sim.C;

	sim.P = (P != -1)? P : std::min(floor((double)Q/n), floor(process_Popt(sim,c)));
	sim.W = (W != -1)? W : process_Topt(sim);

	sim.Hopt = process_Hopt(sim);

	return sim;
}

Simulator createGroupSim(int Q, int n, int k, double P, double W, double d, double c, double v, double mu_s, double alpha)
{
	Simulator sim(n, k, P, W, c, v, c, 0.0, INF, mu_s, alpha);
	sim.C = c + d/((double)Q/n);
	printf("c=%f, d=%f, C=%f\n", c, d, sim.C);
	sim.V = v * sim.C;

	sim.P = 1;
	sim.Pgroup = (P != -1)? P : std::min(floor((double)Q/n), floor(group_Popt(sim,c)));

	sim.mu_s = mu_s / sim.Pgroup;
	sim.W = (W != -1)? W : process_Topt(sim);

	sim.Hopt = process_Hopt(sim);

	return sim;
}

void procExperiment(FILE *file, int Q, int P, double W, int n, int k, double d, double c, double v, double mu_s, double alpha, int nruns, int nw)
{
	Simulator sim = createProcSim(Q, P, W, n, k, d, c, v, mu_s, alpha);
	average(sim, Q, nruns, nw).print(file);
}

void groupExperiment(FILE *file, int Q, int P, double W, int n, int k, double d, double c, double v, double mu_s, double alpha, int nruns, int nw)
{

	Simulator sim = createGroupSim(Q, P, W, n, k, d, c, v, mu_s, alpha);
	average(sim, Q, nruns, nw).print(file);
}

void create(const char *name, FILE **fp22, FILE **fp32, FILE **fg22, FILE **fg32)
{
	char p22[256], p32[256], g22[256], g32[256];
	sprintf(p22, "results/proc/var_%s_n2_k2.out", name);
	sprintf(p32, "results/proc/var_%s_n3_k2.out", name);
	sprintf(g22, "results/group/var_%s_n2_k2.out", name);
	sprintf(g32, "results/group/var_%s_n3_k2.out", name);

	if((*fp22 = fopen(p22, "w")) <= 0) { printf("ERROR\n"); fclose(*fp22); }
	if((*fp32 = fopen(p32, "w")) <= 0) { printf("ERROR\n"); fclose(*fp32); }
	if((*fg22 = fopen(g22, "w")) <= 0) { printf("ERROR\n"); fclose(*fg22); }
	if((*fg32 = fopen(g32, "w")) <= 0) { printf("ERROR\n"); fclose(*fg32); }
}

void close(FILE *fp22, FILE *fp32, FILE *fg22, FILE *fg32)
{
	if(fp22) fclose(fp22);
	if(fp32) fclose(fp32);
	if(fg22) fclose(fg22);
	if(fg32) fclose(fg32);
}

void runall(FILE *fp22, FILE *fp32, FILE *fg22, FILE *fg32,
	int Q, int P, double W, double d, double c, double v, double mu_s, double alpha, int nruns, int nw)
{
	//procExperiment(fp22, Q, 2, 2, P, W, c, C, V, mu_s, alpha, nruns, nw);
	procExperiment(fp32, Q, 3, 2, std::min((Q/3.0),(double)P), W, d, c, v, mu_s, alpha, nruns, nw);
	groupExperiment(fg22, Q, 2, 2, std::min((Q/2.0),(double)P), W, d, c, v, mu_s, alpha, nruns, nw);
	groupExperiment(fg32, Q, 3, 2, std::min((Q/3.0),(double)P), W, d, c, v, mu_s, alpha, nruns, nw);
}

int main(void)
{
	struct timeval time;
	gettimeofday(&time, NULL);
	srand((time.tv_sec * 1000) + (time.tv_usec / 1000));

	int nw = 64;
	int nruns = 30;

	FILE *fp22, *fp32, *fg22, *fg32;

	double Q = 1e6;
	double v = 0;
	double alpha = 1e-8;

	alpha = 1e-6;

	// create("Mu-c1800-a1e6", &fp22, &fp32, &fg22, &fg32);
	// for(double mu=1e6; mu>=1e2; mu/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 1800, v, mu*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// create("Mu-c60-a1e6", &fp22, &fp32, &fg22, &fg32);
	// for(double mu=1e6; mu>=1e2; mu/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 60, v, mu*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// create("Mu-d1e7-a1e6", &fp22, &fp32, &fg22, &fg32);
	// for(double mu=1e6; mu>=1e2; mu/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 1e7, 0, v, mu*Q, alpha, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// create("Mu-c1800-a1e7", &fp22, &fp32, &fg22, &fg32);
	// for(double mu=1e6; mu>=1e2; mu/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 1800, v, mu*Q, 1e-7, nruns, nw);	
	// close(fp22, fp32, fg22, fg32);

	// create("Mu-c1800-a1e5", &fp22, &fp32, &fg22, &fg32);
	// for(double mu=1e6; mu>=1e2; mu/=10)
	// 	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 1800, v, mu*Q, 1e-5, nruns, nw);
	// close(fp22, fp32, fg22, fg32);

	create("Mu2-P-a1e5", &fp22, &fp32, &fg22, &fg32);
	for(double P=10000; P<=500000; P+=10000)
		runall(fp22, fp32, fg22, fg32, Q, P, -1, 0, 1800, v, 1e4*Q, 1e-5, nruns, nw);
	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 1800, v, 1e4*Q, 1e-5, nruns, nw);
	close(fp22, fp32, fg22, fg32);

	create("Mu-P-a1e5", &fp22, &fp32, &fg22, &fg32);
	for(double P=10000; P<=500000; P+=10000)
		runall(fp22, fp32, fg22, fg32, Q, P, -1, 0, 1800, v, 1e3*Q, 1e-5, nruns, nw);
	runall(fp22, fp32, fg22, fg32, Q, -1, -1, 0, 1800, v, 1e3*Q, 1e-5, nruns, nw);
	// close(fp22, fp32, fg22, fg32);

	return 0;
}
